# -*- coding: utf-8 -*-

from skyllh.datasets.i3 import (
    PublicData_10y_ps,
    PublicData_10y_ps_wMC,
    PublicData_10y_ps_wMCEq,
)

data_samples = {
    'PublicData_10y_ps': PublicData_10y_ps,
    'PublicData_10y_ps_wMC': PublicData_10y_ps_wMC,
    'PublicData_10y_ps_wMCEq': PublicData_10y_ps_wMCEq
}
